import abc
import os
import re
import weakref
from pathlib import Path

import inifix
import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import setdefaultattr
from yt.geometry.grid_geometry_handler import GridIndex

from .dmpfile_io import parse_fields_index, read_header, read_idefix_dmpfile
from .fields import IdefixFieldInfo

_IDEFIX_VERSION_REGEXP = re.compile(r"v\d+\.\d+\.?\d*[-\w+]*")


class IdefixGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level, dims):
        super().__init__(id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level
        self.ActiveDimensions = dims

    def __repr__(self):
        return "IdefixGrid_%04i (%s)" % (self.id, self.ActiveDimensions)


class IdefixHierarchy(GridIndex):
    grid = IdefixGrid

    def __init__(self, ds, dataset_type="idefix"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        # float type for the simulation edges and must be float64 now
        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    def _detect_output_fields(self):
        self.field_list = [
            (self.dataset_type, f) for f in self.dataset._detected_field_list
        ]

    def _count_grids(self):
        self.num_grids = 1

    def _parse_index(self):
        self.grid_left_edge[0][:] = self.ds.domain_left_edge[:]
        self.grid_right_edge[0][:] = self.ds.domain_right_edge[:]
        self.grid_dimensions[0][:] = self.ds.domain_dimensions[:]
        self.grid_particle_count[0][0] = 0
        self.grid_levels[0][0] = 1
        self.max_level = 1

        with open(self.ds.parameter_filename, "rb") as fh:
            self._field_offsets = parse_fields_index(fh)

    def _populate_grid_objects(self):
        # the minimal form of this method is
        #
        # for g in self.grids:
        #     g._prepare_grid()
        #     g._setup_dx()
        #
        # This must also set:
        #   g.Children <= list of child grids
        #   g.Parent   <= parent grid
        # This is handled by the frontend because often the children must be identified.
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            g = self.grid(i, self, self.grid_levels.flat[i], self.grid_dimensions[i])
            g._prepare_grid()
            g._setup_dx()
            self.grids[i] = g


class IdefixDataset(Dataset, abc.ABC):
    pass


class IdefixDumpDataset(IdefixDataset):
    _index_class = IdefixHierarchy
    _field_info_class = IdefixFieldInfo

    def __init__(
        self,
        dmpfile,
        inifile=None,
        dataset_type="idefix",
        unit_system="cgs",
        units_override=None,
    ):
        self.fluid_types += ("idefix",)
        self.inifile = inifile
        super().__init__(
            dmpfile,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = None

        # idefix does not support grid refinement
        self.refine_by = 1

    def _set_code_unit_attributes(self):
        # This is where quantities are created that represent the various
        # on-disk units.  These are the currently available quantities which
        # should be set, along with examples of how to set them to standard
        # values.
        #
        # self.length_unit = self.quan(1.0, "cm")
        # self.mass_unit = self.quan(1.0, "g")
        # self.time_unit = self.quan(1.0, "s")
        # self.time_unit = self.quan(1.0, "s")
        #
        # These can also be set:
        # self.velocity_unit = self.quan(1.0, "cm/s")
        # self.magnetic_unit = self.quan(1.0, "gauss")
        for key, unit in self.__class__.default_units.items():
            setdefaultattr(self, key, self.quan(1, unit))

    def _parse_parameter_file(self):

        # first pass in the dmpfile: read everything except large arrays
        fprops, fdata = read_idefix_dmpfile(self.parameter_filename, skip_data=True)
        self._detected_field_list = [k for k in fprops if re.match(r"^V[sc]-", k)]

        # parse the version hash
        header = read_header(self.parameter_filename)

        match = re.search(_IDEFIX_VERSION_REGEXP, header)
        self.parameters["idefix version"] = (
            "unknown" if match is None else match.group()
        )

        if self.inifile is not None:
            self.parameters.update(inifix.load(self.inifile))
            grid_ini = self.parameters["Grid"]
            for ax, vals in grid_ini.items():
                if vals[0] > 1:
                    # more than one block is only relevant for mixing grid spacings,
                    # but only "u" is supported
                    raise ValueError(f"Unsupported block structure for {ax}.")
                if vals[3] != "u":
                    raise ValueError(f"Unsupported grid spacing '{vals[3]}'.")

        # parse the grid
        axes = ("x1", "x2", "x3")
        self.domain_dimensions = np.concatenate([fprops[k][-1] for k in axes])
        self.dimensionality = np.count_nonzero(self.domain_dimensions - 1)

        # note that domain edges parsing is already implemented in a mutli-block
        # supporting fashion even though we specifically error out in case there's more
        # than one block.
        self.domain_left_edge = np.array(
            [fdata[f"xl{idir}"][0] for idir in "123"], dtype="float64"
        )
        self.domain_right_edge = np.array(
            [fdata[f"xr{idir}"][-1] for idir in "123"], dtype="float64"
        )

        self.current_time = fdata["time"]

        self._periodicity = tuple(bool(p) for p in fdata["periodicity"])
        enum_geoms = {1: "cartesian", 2: "cylindrial", 3: "polar", 4: "spherical"}
        self.geometry = enum_geoms[fdata["geometry"]]

        # idefix is never cosmological
        self.cosmological_simulation = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0

    @classmethod
    def _is_valid(self, fn, *args, **kwargs):
        ok = bool(re.match(r"^(dump)\.\d{4}(\.dmp)$", Path(fn).name))
        try:
            ok &= "idefix" in read_header(fn).lower()
        except Exception:
            ok = False
        return ok



# register our dataset classes manually
from yt.utilities.object_registries import output_type_registry

from .data_structures import IdefixDumpDataset, IdefixGrid, IdefixHierarchy
from .fields import IdefixFieldInfo
from .io import IdefixIOHandler

output_type_registry["IdefixDumpDataset"] = IdefixDumpDataset


__version__ = "0.1.0"



import numpy as np

from yt.utilities.io_handler import BaseIOHandler

from .dmpfile_io import read_single_field


class IdefixIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "idefix"

    def __init__(self, ds):
        BaseIOHandler.__init__(self, ds)
        self.ds = ds
        self.dmpfile = ds.parameter_filename

    def _read_fluid_selection(self, chunks, selector, fields, size):
        data = {}
        offset = 0

        with open(self.dmpfile, mode="rb") as fh:
            for field in fields:
                ftype, fname = field
                data[ftype, fname] = np.empty(size, dtype="float64")
                for chunk in chunks:
                    for grid in chunk.objs:
                        foffset = grid._index._field_offsets[fname]
                        values = read_single_field(fh, foffset)
                        offset += grid.select(selector, values, data[field], offset)
        return data

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        raise NotImplementedError

    def _read_particle_fields(self, chunks, ptf, selector):
        # idefix doesn't have particles (yet)
        raise NotImplementedError

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        raise NotImplementedError



import re
import struct
from typing import List

import numpy as np

# hardcoded in idefix
HEADERSIZE = 128
NAMESIZE = 16

SIZE_CHAR = 1
SIZE_INT = 4
# emulating C++
# enum DataType {DoubleType, SingleType, IntegerType};
DTYPES = ["d", "f", "i"]


def read_null_terminated_string(fh, maxsize=NAMESIZE):
    """Read maxsize * SIZE_CHAR bytes, but only parse non-null characters."""
    b = fh.read(maxsize * SIZE_CHAR)
    s = b.decode("utf-8", errors="backslashreplace")
    s = s.split("\x00", maxsplit=1)[0]
    return s


def read_next_field_properties(fh):
    """Emulate Idefix's OutputDump::ReadNextFieldProperty"""
    field_name = read_null_terminated_string(fh)

    fmt = "=i"
    dtype = DTYPES[struct.unpack(fmt, fh.read(struct.calcsize(fmt)))[0]]
    ndim = struct.unpack(fmt, fh.read(struct.calcsize(fmt)))[0]
    if ndim > 3:
        raise ValueError(ndim)
    fmt = f"={ndim}i"
    dim = np.array(struct.unpack(fmt, fh.read(struct.calcsize(fmt))))
    return field_name, dtype, ndim, dim


def read_chunk(fh, ndim: int, dim: List[int], dtype, is_scalar=False, skip_data=False):
    assert ndim == len(dim)
    fmt = f"={np.product(dim)}{dtype}"
    size = struct.calcsize(fmt)
    if skip_data:
        fh.seek(size, 1)
        return None
    data = struct.unpack(fmt, fh.read(size))

    # note: this reversal may not be desirable in general
    if is_scalar:
        return data[0]

    data = np.reshape(data, dim, order="F")
    return data


def read_serial(fh, ndim: int, dim: List[int], dtype, is_scalar=False):
    """Emulate Idefix's OutputDump::ReadSerial"""
    assert ndim == 1  # corresponds to an error raised in IDEFIX
    return read_chunk(fh, ndim=ndim, dim=dim, dtype=dtype, is_scalar=is_scalar)


def read_distributed(fh, dim, skip_data=False):
    """Emulate Idefix's OutputDump::ReadDistributed"""
    # note: OutputDump::ReadDistributed only read doubles
    # because chucks written in integers are small enough
    # that parallelization is counter productive.
    # This a design choice on idefix's size.
    return read_chunk(fh, ndim=len(dim), dim=dim, dtype="d", skip_data=skip_data)


# The following functions are originally designed for yt


def read_header(filename):
    with open(filename, "rb") as fh:
        header = read_null_terminated_string(fh, maxsize=HEADERSIZE)
    return header


def parse_fields_index(fh):
    """
    Go over a dumpfile, parse bytes offsets associated with each field.
    """
    field_index = {}

    # skip header
    fh.seek(HEADERSIZE)
    # skip grid properties
    for _ in range(9):
        _field_name, dtype, ndim, dim = read_next_field_properties(fh)
        read_serial(fh, ndim, dim, dtype)

    while True:
        offset = fh.tell()
        field_name, dtype, ndim, dim = read_next_field_properties(fh)
        if not re.match("^V[cs]-", field_name):
            break
        field_index[field_name] = offset
        read_distributed(fh, dim, skip_data=True)

    return field_index


def read_single_field(fh, field_offset):
    fh.seek(field_offset)
    field_name, dtype, ndim, dim = read_next_field_properties(fh)
    data = read_distributed(fh, dim)
    return data


def read_idefix_dmpfile(filename, skip_data=False):
    with open(filename, "rb") as fh:
        return read_idefix_dump_from_buffer(fh, skip_data)


def read_idefix_dump_from_buffer(fh, skip_data=False):

    # skip header
    fh.seek(HEADERSIZE)

    fprops = {}
    fdata = {}
    for _ in range(9):
        # read grid properties
        # (cell centers, left and right edges in 3D -> 9 arrays)
        field_name, dtype, ndim, dim = read_next_field_properties(fh)
        data = read_serial(fh, ndim, dim, dtype)
        fprops[field_name] = dtype, ndim, dim
        fdata[field_name] = data

    field_name, dtype, ndim, dim = read_next_field_properties(fh)
    while field_name != "eof":
        # note that this could likely be implemented using a call to
        # `iter` with a sentinel value, to the condition that read_next_field_properties
        # would be splitted into 2 parts (I don't the sentinel pattern works with tuples)
        fprops[field_name] = dtype, ndim, dim
        if field_name.startswith("Vc-") or field_name.startswith("Vs-"):
            data = read_distributed(fh, dim, skip_data=skip_data)
        else:
            is_scalar = ndim == 1 and dim[0] == 1
            is_scalar &= field_name not in ("x1", "x2", "x3")
            data = read_serial(fh, ndim, dim, dtype, is_scalar=is_scalar)
        fdata[field_name] = data
        field_name, dtype, ndim, dim = read_next_field_properties(fh)

    return fprops, fdata
